import json
import os
import sys
import shutil

import pytest
from idmtools.analysis.analyze_manager import AnalyzeManager
from idmtools.analysis.download_analyzer import DownloadAnalyzer
from idmtools.assets import Asset
from idmtools.builders import SimulationBuilder
from idmtools.core import ItemType
from idmtools.core.platform_factory import Platform
from idmtools.entities.experiment import Experiment
from idmtools.entities.simulation import Simulation
from idmtools_test.utils.itest_with_persistence import ITestWithPersistence
from emodpy.emod_task import EMODTask
from emodpy.generic.serialization import add_serialization_timesteps, load_serialized_population
from emodpy.utils import download_latest_bamboo, download_latest_schema, EradicationBambooBuilds, bamboo_api_login

from examples.config_update_parameters import del_folder
from emod_api.config import from_schema as fs
from emod_api.schema import get_schema as gs

from . import manifest

eradication_path = manifest.schema_path_linux

current_directory = os.path.dirname(os.path.realpath(__file__))
BIN_PATH = os.path.join(current_directory, "..", "examples", "inputs", "bin")
INPUT_PATH = os.path.join(current_directory, "..", "examples", "serialization", "inputs")


@pytest.mark.smoke
@pytest.mark.comps  # Change to @pytest.mark.emod when serialization is ready
class TestSerialization(ITestWithPersistence):

    @classmethod
    def setUpClass(cls) -> None:
        cls.platform = Platform("SLURM")
        cls.plan = EradicationBambooBuilds.CI_GENERICLINUX
        cls.schema_bamboo_path = manifest.schema_path_linux
        if not os.path.isfile(eradication_path):
            bamboo_api_login()
            download_latest_bamboo(plan=cls.plan, scheduled_builds_only=False, out_path=eradication_path)

    def setUp(self) -> None:
        self.case_name = os.path.basename(__file__) + "--" + self._testMethodName
        print(self.case_name)
        self.schema_file = os.path.join(manifest.schema_folder, 'schema.json')
        self.config_file = os.path.join(manifest.config_folder, 'config.json')
        manifest.delete_existing_file(self.schema_file)
        manifest.delete_existing_file(self.config_file)
        if os.path.isfile(self.schema_bamboo_path):
            shutil.copy(self.schema_bamboo_path, self.schema_file)
        else:
            if sys.platform == "linux" or sys.platform == "linux2":
                # try to generate schema from eradication
                print(
                    f"Your OS is {sys.platform}, trying to generate schema from eradication {eradication_path}")
                gs.dtk_to_schema(eradication_path, path_to_write_schema=self.schema_file)
            else:
                print(
                    f"Your OS is {sys.platform}, trying to use download schema from bamboo.")
                bamboo_api_login()
                download_latest_schema(plan=self.plan, scheduled_builds_only=False, out_path=self.schema_file)
        # generate default config file from schema
        fs.SchemaConfigBuilder(schema_name=self.schema_file, config_out=self.config_file)

    @pytest.mark.long
    def test_serialization(self):
        # Step1: create experiment and simulations with serialization file in output
        sim_duration = 1  # in years
        serialization_timestep = 30
        num_seeds = 1

        def set_param_fn(config):
            config.parameters.Enable_Demographics_Reporting = 1
            config.parameters.Incubation_Period_Distribution = "CONSTANT_DISTRIBUTION"
            config.parameters.Incubation_Period_Constant = 2
            config.parameters.Infectious_Period_Distribution = "CONSTANT_DISTRIBUTION"
            config.parameters.Infectious_Period_Constant = 3
            config.parameters.Base_Infectivity_Distribution = "CONSTANT_DISTRIBUTION"
            config.parameters.Base_Infectivity_Constant = 0.2
            config.parameters.Post_Infection_Acquisition_Multiplier = 0.7
            config.parameters.Post_Infection_Transmission_Multiplier = 0.4
            config.parameters.Post_Infection_Mortality_Multiplier = 0.3
            config.parameters.Simulation_Duration = sim_duration * 365
            return config

        print(f"The config.json should be in {self.config_file}.")

        from emod_api.interventions import outbreak as ob
        from emod_api import campaign as camp

        print(f"Telling emod-api to use {self.schema_file} as schema.")
        ob.schema_path = self.schema_file

        # ToDo: when #276 is resolved, change campaign_builder=build_camp to campaign_builder=camp
        def build_camp():
            event = ob.new_intervention(1, cases=4)
            camp.add(event)
            # We are saving and reloading. Maybe there's an even better way? But even an outbreak seeding does not belong in the EMODTask.
            # camp_path = "campaign.json"
            # camp.save(camp_path)
            return camp

        task1 = EMODTask.from_default2(config_path=self.config_file, eradication_path=eradication_path,
                                       campaign_builder=build_camp, schema_path=self.schema_file, param_custom_cb=set_param_fn)
        timesteps = [serialization_timestep]
        add_serialization_timesteps(task=task1, timesteps=timesteps,
                                    end_at_final=False, use_absolute_times=False)

        builder = SimulationBuilder()
        builder.add_sweep_definition(task1.set_parameter_partial("Run_Number"), range(num_seeds))

        experiment1 = Experiment.from_builder(builder, task1, name=self.case_name + " create serialization")

        self.platform.run_items(experiment1)
        self.platform.wait_till_done(experiment1)
        self.assertTrue(experiment1.succeeded)

        # download serialized file for later reload similation

        # Cleanup the output path
        output_path = 'output'
        del_folder(output_path)

        # We want to download all the dtk state files and the InsetChart.json
        filenames = []
        for serialization_timestep in timesteps:
            filenames.append("output/state-" + str(serialization_timestep).zfill(5) + ".dtk")
        filenames.append('output/InsetChart.json')

        analyzers = [DownloadAnalyzer(filenames=filenames)]
        # self.generate_experiment()
        am = AnalyzeManager(platform=self.platform, ids=[(experiment1.id, ItemType.EXPERIMENT)], analyzers=analyzers)
        am.analyze()

        # Step2: Create new experiment and sim with previous serialized file
        task2 = EMODTask.from_default2(config_path=self.config_file, eradication_path=eradication_path,
                                       campaign_builder=build_camp, schema_path=self.schema_file,
                                       param_custom_cb=set_param_fn)
        # task2 = EMODTask.from_default(default=EMODSir(), eradication_path=os.path.join(BIN_PATH, 'Eradication.exe'))
        reload_sim = Simulation.from_task(task2)
        # Add the serialized population file to the experiment
        serialized_population_file = Asset(
            absolute_path=os.path.join('output', experiment1.simulations[0].id,
                                       "state-" + str(serialization_timestep).zfill(5) + ".dtk"))
        task2.common_assets.add_asset(serialized_population_file)

        # Serialization parameters: Enable serialization and reload the population
        reload_sim.task.set_parameter("Simulation_Duration", sim_duration * 365)
        state_file = 'state-' + str(serialization_timestep).zfill(5) + '.dtk'
        load_serialized_population(task2, population_path="Assets",
                                   population_filenames=[state_file])
        experiment2 = Experiment(name=self.case_name + " realod serialization")
        experiment2.tags = {'idmtools': 'reload serialization', 'exp2': 'tag2'}
        experiment2.simulations.append(reload_sim)
        self.platform.run_items(experiment2)
        self.platform.wait_till_done(experiment2)
        self.assertTrue(experiment2.succeeded)

        # validate InsetChart.json from experiment2 has timestep = sim_duration * 365
        for sim in experiment2.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            time_steps = json.loads(files['output/InsetChart.json'])['Header']['Timesteps']
            self.assertEqual(sim_duration * 365, time_steps)
        # todo: add analyzer to verify the serialization file is loaded correctly.
